from .ellatu_db import Document, Model, MongoId, UserKey
from typing import Dict, List, Optional, Text
from .ellatu import (MessageSegment, Request, TextMessage, add_msg, data_action,
                     limit_codeblocks, limit_columns, limit_lines, limit_users,
                     remove_files, terminate_request, trace, pipeline_sequence,
                     RequestAction, EllatuPipeline, MessageType,
                     ParagraphMessage)
from . import mapper

PARSER = mapper.mapper_parser()


DEFAULT_SETTINGS: Dict[str, Optional[int]] = {
    "users": None,
    "blocks": 3,
    "lines": 10,
    "cols": 79
}


def compile_codeblocks(parser) -> RequestAction:
    def action(request: Request) -> Request:
        models: Dict[MongoId, List[Model]] = {}

        for userid, codeblocks in request.codeblocks.items():
            try:
                models[userid] = mapper.compile_codeblocks(codeblocks, parser)
            except mapper.MapperError as e:
                return terminate_request(request, "Compilation error:\n " +
                                         str(e))

        request.data["models"] = models
        return trace(request, "Success")
    return action


def _requester_owns_main(request: Request,
                         models: Dict[MongoId, mapper.Model]) -> bool:
    models = request.data['models']
    requestor_models = models[request.users[request.requestor]['_id']]
    for model in requestor_models:
        if mapper.contains_main(model):
            return True
    return False


def _create_model_list(request: Request,
                       models: Dict[MongoId, mapper.Model],
                       blocks_order: List[mapper.Model]) -> List[mapper.Model]:
    requestor_models = models[request.users[request.requestor]['_id']]
    m_list = []
    for userkey in blocks_order:
        if userkey == request.requestor or userkey not in request.users:
            continue
        if request.users[userkey]['_id'] not in models:
            continue
        users_models = models[request.users[userkey]['_id']]
        m_list += users_models
    m_list += requestor_models
    return m_list


@data_action(["models", "blocks_order"])
def run_models(request: Request,
               models: Dict[MongoId, List[mapper.Model]],
               blocks_order: List[UserKey]) -> Request:
    if request.level is None:
        return terminate_request(request, "Invalid level")
    if request.requestor not in request.users:
        return terminate_request(request, "Requestor not added to the " +
                                 "submission")
    if request.users[request.requestor]['_id'] not in models:
        return terminate_request(request, "Requestor doesn't have compiled " +
                                 "code in the submission")

    if not _requester_owns_main(request, models):
        return terminate_request(request,
                                 "Requestor doesn't own a main function")

    try:
        model_list = _create_model_list(request, models, blocks_order)
        _, mapp, ship, out = mapper.run_models(model_list)
    except mapper.MapperError as e:
        return terminate_request(request, "Mapper error: " + str(e))

    if out:
        request.add_message(TextMessage(f"_Output:_\n" + '\n'.join(out)))

    for test in request.level['tests']:
        print(test)
        desired_map = mapper.fill_from_text(mapper.Map(), test, ';')
        diff_map = mapper.mapp_diff(desired_map, mapp)
        if diff_map is not None:
            filename = request.ellatu.temp_files.add_temp_filename(
                str(request.id) + '-mapperdiff.png'
            )
            mapper.render_map(filename, diff_map, ship)
            request.add_message(
                ParagraphMessage(f"here is the diff ![diff][{filename}]",
                                 message_type=MessageType.ERROR))
            request.alive = False
            request.add_on_res(remove_files([filename]))
            return request

    return trace(request, "Passed all tests")


def get_level_settings(level: Document) -> Dict[str, Optional[int]]:
    settings = dict(DEFAULT_SETTINGS)
    for key, value in level["attrs"].items():
        settings[key] = value
    return settings


def check_blocks(request: Request) -> Request:
    if request.level is None:
        return terminate_request(request, "The level is not set")

    settings = get_level_settings(request.level)
    return pipeline_sequence([
        limit_users(settings["users"]),
        limit_codeblocks(settings["blocks"]),
        limit_lines(settings["lines"]),
        limit_columns(settings["cols"])
    ])(request)


class MapperPipeline(EllatuPipeline):

    def __init__(self) -> None:
        self.parser = mapper.mapper_parser()

    def on_submit(self) -> RequestAction:
        return pipeline_sequence([
            check_blocks,
            add_msg(MessageSegment('Compilation')),
            compile_codeblocks(self.parser)
        ])

    def on_run(self) -> RequestAction:
        return pipeline_sequence([
            self.on_submit(),
            add_msg(MessageSegment('Testing')),
            run_models
        ])



from collections import defaultdict
from .document_editor import DocumentEditor
from typing import (Any, Callable, DefaultDict, Generator, List, Optional,
                    Tuple, Dict, TypeVar, Union)
from PIL import Image, ImageDraw
import numpy as np
from math import pi
import tatsu
from tatsu.walkers import NodeWalker
import sys
import os

from importlib import resources

Color = str
TileColor = Tuple[Optional[Color], Optional[Color]]

T = TypeVar("T")
R = TypeVar("R")
S = TypeVar("S")

###############################################################################
# Error
###############################################################################


class MapperError(Exception):
    pass


class MapperRuntimeError(MapperError):
    pass


class MapperCompilationError(MapperError):
    pass

###############################################################################
# TUPLE OPERATIONS
###############################################################################


def ewise_tuple(a: Tuple[T, ...], b: Tuple[R, ...],
                f: Callable[[T, R], S]) -> Tuple[S, ...]:
    return tuple(f(ae, be) for ae, be in zip(a, b))


def tcut2(a: Tuple[int, ...]) -> Tuple[int, int]:
    return a[0], a[1]


def tadd(a: Tuple[int, ...], b: Tuple[int, ...]) -> Tuple[int, ...]:
    return ewise_tuple(a, b, lambda x, y: x + y)


def tsub(a: Tuple[int, ...], b: Tuple[int, ...]) -> Tuple[int, ...]:
    return ewise_tuple(a, b, lambda x, y: x - y)


def tmul(a: Tuple[int, ...], b: Tuple[int, ...]) -> Tuple[int, ...]:
    return ewise_tuple(a, b, lambda x, y: x * y)


def tdiv(a: Tuple[int, ...], b: Tuple[int, ...]) -> Tuple[float, ...]:
    return ewise_tuple(a, b, lambda x, y: x / y)


def tint(a: Tuple[float, ...]) -> Tuple[int, ...]:
    return tuple(int(v) for v in a)


def tran_position(pos: Tuple[int, int], dims: Tuple[int, int],
                  tile_width: int, center: Tuple[float, float]) \
        -> Tuple[float, float]:
    x = ((pos[0] - center[0]) * tile_width + dims[0] // 2)
    y = ((pos[1] - center[1]) * tile_width + dims[1] // 2)
    return x, y

###############################################################################
# MAP AND SHIP
###############################################################################


class Map:

    def __init__(self, start: Tuple[int, int] = (0, 0)) -> None:
        self.start = start
        self.tiles: DefaultDict[Tuple[int, int], TileColor] = \
            defaultdict(lambda: (None, None))

    def get_bounding_coords(self) -> Tuple[int, int, int, int]:

        min_x, min_y = self.start
        max_x, max_y = self.start

        for (x, y), _ in self.tiles.items():
            min_x, min_y = min(min_x, x), min(min_y, y)
            max_x, max_y = max(max_x, x), max(max_y, y)

        return min_x, min_y, max_x, max_y

    def get(self, tile: Tuple[int, int]) -> TileColor:
        return self.tiles[tile]

    def put(self, tile: Tuple[int, int],
            color: Optional[Color] = None,
            border_color: Optional[Color] = None) -> None:

        (new_color, new_border) = self.tiles[tile]
        if color is not None:
            new_color = color
        if border_color is not None:
            new_border = border_color
        self.tiles[tile] = (new_color, new_border)

    def get_tiles(self) -> Generator[Tuple[Tuple[int, int], TileColor],
                                     None, None]:
        yield from self.tiles.items()
        yield self.start, (None, 'green')

    def reset(self) -> None:
        self.tiles.clear()

    def shift(self, dx: int, dy: int) -> None:
        temp_tiles = {}
        for (x, y), color in self.tiles.items():
            temp_tiles[(x + dx, y + dy)] = color
        self.tiles.clear()

        for tile, color in temp_tiles.items():
            self.tiles[tile] = color

    def copy(self) -> 'Map':
        result = Map(start=self.start)
        for key, color in self.tiles.items():
            result.tiles[key] = color
        return result


def mapp_diff(desired: Map, received: Map) -> Optional[Map]:
    result = received.copy()

    different = False

    ana_from, ana_on = desired, received
    miss_color = 'red'
    diff_color = 'yellow'
    for _ in range(2):
        for tile, (color, _) in ana_from.tiles.items():
            if color is None:
                continue
            (rec_color, _) = ana_on.get(tile)
            if rec_color is None:
                result.put(tile, color=None, border_color=miss_color)
                different = True
                continue
            if rec_color != color:
                result.put(tile, color=None, border_color=diff_color)
                different = True
        ana_from, ana_on = ana_on, ana_from
        miss_color = 'blue'

    return result if different else None


def mov_for_dir(direction: int) -> Tuple[int, int]:
    assert 0 <= direction <= 3, "There are only four directions"
    if direction == 0:
        return (0, -1)
    if direction == 1:
        return (1, 0)
    if direction == 2:
        return (0, 1)
    if direction == 3:
        return (-1, 0)
    return (0, 0)


class Ship:

    def __init__(self, mapp: Map, position: Tuple[int, int] = (0, 0)) -> None:
        self.position = position
        self.direction = 0
        self.mapp = mapp
        self.pendown = True
        self.color = 'black'

    def move_fwd(self, amount: int) -> 'Ship':
        if self.pendown:
            self.put(self.color)

        direction = self.direction
        if amount < 0:
            direction = (direction + 2) % 4

        for _ in range(abs(amount)):
            self.move(mov_for_dir(direction))

            if self.pendown:
                self.put(self.color)

        return self

    def move(self, movement: Tuple[int, int]) -> 'Ship':
        self.position = tcut2(tadd(self.position, movement))
        return self

    def rotate(self, amount: int) -> 'Ship':
        self.direction = (self.direction + amount) % 4
        return self

    def put(self, color: Optional[Color] = None,
            border_color: Optional[Color] = None) -> None:
        self.mapp.put(self.position, color=color, border_color=border_color)


COLORS: Dict[str, Optional[Color]] = {
    'b': "black",
    'g': "green",
    'l': "blue",
    'r': "red",
    '_': None
}

COLORS_BACK: Dict[Color, str] = {}
for char, color in COLORS.items():
    if color:
        COLORS_BACK[color] = char


def get_color(character: str) -> Optional[Color]:
    if character not in COLORS:
        return None
    return COLORS[character]


def fill_from_text_list(mapp: Map, lines: List[str]) -> Map:
    for line in lines:
        col, row, color = line.split()
        mapp.tiles[(int(col), int(row))] = (COLORS[color], None)
    return mapp


def fill_from_text_map(mapp: Map, lines: List[str]) -> Map:
    cx, cy = 0, 0
    for row, line in enumerate(lines):
        for col, ch in enumerate(line):
            color = get_color(ch.lower())
            if ch == 'S' or ch.isupper():
                cx, cy = (col, row)
            mapp.tiles[(col, row)] = (color, None)
    mapp.shift(-cx, -cy)
    return mapp


def fill_from_text(mapp: Map, text: str, sep: Optional[str] = None) -> Map:
    lines = text.splitlines() if sep is None else text.split(sep)
    header = lines[0].strip()
    if header == 'LIST':
        return fill_from_text_list(mapp, lines[1:])  # TODO: yeah slice bad
    return fill_from_text_map(mapp, lines)


def fill_from_file(mapp: Map, path: str, sep: Optional[str] = None) -> Map:
    with open(path, "r") as f:
        return(fill_from_text(mapp, f.read(), sep))


def export_map(mapp: Map, sep: str = '\n') -> str:
    res = 'LIST'
    for (col, row), (color, _) in mapp.tiles.items():
        if color is not None:
            res += sep + f'{col} {row} {COLORS_BACK[color]}'
    return res


###############################################################################
# TRANSFORM
###############################################################################


def rotation_matrix(rot: float) -> np.ndarray:
    c, s = np.cos(rot), np.sin(rot)
    return np.array(((c, -s, 0), (s, c, 0), (0, 0, 1)))


def translation_matrix(pos: Tuple[int, int]) -> np.ndarray:
    return np.array(((1, 0, pos[0]), (0, 1, pos[1]), (0, 0, 1)))


def scale_matrix(scale: float) -> np.ndarray:
    return np.array(((scale, 0, 0), (0, scale, 0), (0, 0, 1)))


def transform(polygon: np.ndarray, pos: Tuple[int, int],
              rot: float, scale: float) -> np.ndarray:

    t, r, s = translation_matrix(
        pos), rotation_matrix(rot), scale_matrix(scale)
    return t.dot(r.dot(s.dot(polygon)))

###############################################################################
# RENDERING
###############################################################################


SHIP = np.array(((0, -1, 1), (1, 1, 1), (-1, 1, 1))).T


def render_map(path: str, mapp: Map, ship: Optional[Ship] = None) -> None:
    im = draw_map(mapp, ship)
    im.save(path)


def show_map(mapp: Map, ship: Optional[Ship] = None) -> None:
    im = draw_map(mapp, ship)
    im.show()


def draw_grid(g: ImageDraw.ImageDraw, box: Tuple[int, int, int, int],
              dims: Tuple[int, int], center: Tuple[float, float],
              tile_size: int, color: str = 'lightgray', width: int = 2,
              bold_color: str = 'gray') -> None:

    coord = tran_position((0, 0), dims, tile_size, center)

    width, height = dims
    left, top, right, bottom = box
    ln_off_x = width / 2 - (width / 2 // tile_size * tile_size)
    if (right-left) % 2 == 0:
        ln_off_x -= tile_size / 2
    ln_off_y = height / 2 - (height / 2 // tile_size * tile_size)
    if (top - bottom) % 2 == 0:
        ln_off_y += tile_size / 2

    ln_off_x = coord[0] - tile_size / 2
    ln_off_y = coord[1] + tile_size / 2

    off_col = coord[0] // tile_size
    off_row = coord[1] // tile_size

    frequencies = [(1, color), (5, bold_color)]
    for freq, line_col in frequencies:
        for col in range(0, width // tile_size + 1):
            if (col - off_col) % freq == 0:
                x = (col - off_col) * tile_size + ln_off_x
                g.line((x, 0, x, height), fill=line_col, width=2)
        for row in range(0, height // tile_size + 1):
            if (row - off_row) % freq == 0:
                y = (row - off_row) * tile_size + ln_off_y
                g.line((0, y, width, y), fill=line_col, width=2)


def draw_map(mapp: Map, ship: Optional[Ship] = None) -> Image.Image:
    width = 800
    height = 500

    left, top, right, bottom = mapp.get_bounding_coords()
    cx, cy = (left + right) / 2, (top + bottom) / 2

    tile_width = 50

    if right != cx:
        tile_width = int(min(tile_width, width / 2 / (right - cx + 1)))
    if left != cx:
        tile_width = int(min(tile_width, width / 2 / (cx - left + 1)))
    if bottom != cy:
        tile_width = int(min(tile_width, height / 2 / (bottom - cy + 1)))
    if top != cy:
        tile_width = int(min(tile_width, height / 2 / (cy - top + 1)))

    tile_width = int(tile_width)

    dims = (width, height)
    center = (cx, cy)

    im = Image.new(mode="RGB", size=dims, color="white")
    g = ImageDraw.Draw(im)

    draw_grid(g, (left, top, right, bottom), dims, center,
              tile_width, color='#eeeeee', bold_color='#bbbbbb')

    for pos, (color, border_color) in mapp.get_tiles():
        coord = tran_position(pos, dims, tile_width, center)
        coord = coord[0] - tile_width // 2, coord[1] - tile_width // 2
        g.rectangle(
            [coord, (coord[0] + tile_width, coord[1] + tile_width)],
            fill=color, width=tile_width//10,
            outline=border_color)

    if ship is not None:
        rot = ship.direction * pi / 2
        ship_pos = tran_position(ship.position, dims, tile_width, center)
        polygon = transform(SHIP, (int(ship_pos[0]), int(ship_pos[1])), rot,
                            tile_width / 3).T.tolist()
        res_poly = tuple([tuple(p[:2]) for p in polygon])
        g.polygon(tuple(res_poly), "gray")

    return im


###############################################################################
# PARSING
###############################################################################

MapperParser = Any


def mapper_parser(asmodel: bool = True) -> MapperParser:
    with resources.open_text('ellatu', 'mapper.ebnf') as f:
        grammar = f.read()
        return tatsu.compile(grammar, asmodel=asmodel)


def get_indentation(line: str) -> int:
    res = 0
    for ch in line:
        if ch != ' ':
            break
        res += 1
    return res


def is_empty(line: str) -> bool:
    return len(line.strip()) == 0


def indent_to_codeblocks(code: str) -> str:
    lines = code.splitlines(False)

    res = ""

    indents = [0]

    for line in lines:
        if is_empty(line):
            continue
        cur_indent = get_indentation(line)

        if cur_indent > indents[-1]:
            res += ' ' * indents[-1] + '{\n'
            indents.append(cur_indent)
        while cur_indent < indents[-1]:
            indents.pop()
            res += ' ' * indents[-1] + '}\n'

        res += line + '\n'

    for _ in range(len(indents) - 1):
        res += '}\n'

    return res


def mapper_preprocessor(code: str) -> str:
    return indent_to_codeblocks(code)


class ScopeStack:

    def __init__(self, stack: Optional[List[Dict[str, Any]]] = None):
        self.stack: List[Dict[str, Any]] = stack if stack is not None else []

    def lookup(self, symbol: str) -> Optional[Any]:
        for dic in reversed(self.stack):
            if symbol in dic:
                return dic[symbol]
        return None

    def put(self, symbol: str, value: Any, min_layer: int = 0) -> None:
        assert self.stack

        for index in range(len(self.stack) - 1, min_layer - 1, -1):
            if symbol in self.stack[index]:
                self.stack[index][symbol] = value
                return

        self.stack[-1][symbol] = value

    def put_on_last(self, symbol: str, value: Any) -> None:
        assert self.stack
        self.stack[-1][symbol] = value

    def add_scope(self) -> None:
        self.stack.append({})

    def pop_scope(self) -> None:
        self.stack.pop()

    def copy(self, layers: int = 1) -> 'ScopeStack':
        return ScopeStack(self.stack[:layers])


ARIT_OPS: Dict[str, Callable[[int, int], int]] = {
    '&&': lambda a, b: int(a and b),
    '||': lambda a, b: int(a or b),
    '=>': lambda a, b: int(a and not b),

    '<': lambda a, b: int(a < b),
    '<=': lambda a, b: int(a <= b),
    '>': lambda a, b: int(a > b),
    '>=': lambda a, b: int(a >= b),

    '==': lambda a, b: int(a == b),
    '!=': lambda a, b: int(a != b),

    '+': lambda a, b: a + b,
    '-': lambda a, b: a - b,
    '*': lambda a, b: a * b,
    '/': lambda a, b: a // b,
    '%': lambda a, b: a % b,
}

MapperASTNode = Any

MapperInbuiltFunction = Callable[['MapperWalker', List[Any]], Any]


class MapperInbuiltFunctionWrapper():
    def __init__(self, function: MapperInbuiltFunction):
        self.function = function


class MapperWalker(NodeWalker):

    def __init__(self, glob_scope: ScopeStack,
                 inbuilt: Optional[
                     Dict[str, Union[MapperInbuiltFunction, int]]],
                 ship: Ship,
                 mapp: Map) \
            -> None:
        self.ctx = glob_scope
        self.scope = glob_scope
        if inbuilt:
            for key, value in inbuilt.items():
                if isinstance(value, int):
                    self.scope.put(key, value)
                else:
                    self.scope.put(key, MapperInbuiltFunctionWrapper(value))
        self.ship = ship
        self.mapp = mapp
        self.max_out = 50
        self.out: List[str] = []

    def walk_object(self, node: MapperASTNode) -> MapperASTNode:
        return node

    def walk__function(self, node: MapperASTNode) -> MapperASTNode:
        self.scope.put(node.name, node)
        return node

    def walk__amn_function(self, node: MapperASTNode) -> MapperASTNode:
        self.walk(node.fun)
        return None

    def walk__codeblock(self, node: MapperASTNode) -> MapperASTNode:
        self.scope.add_scope()
        value = None
        for line in node.lines:
            value = self.walk(line)
            if value is not None:
                break
        self.scope.pop_scope()
        return value

    def walk__if_statement(self, node: MapperASTNode) -> MapperASTNode:
        # print('if')

        if self.walk(node.cond):
            return self.walk(node.body)

        for elif_stmt in node.elifs:
            if self.walk(elif_stmt.cond):
                return self.walk(elif_stmt.body)

        if node.else_block is not None:
            return self.walk(node.else_block)

        return None

    def walk__while_stmt(self, node: MapperASTNode) -> MapperASTNode:

        while self.walk(node.cond):
            result = self.walk(node.body)
            if result is not None:
                lift, value = result
                if lift == 'break':
                    if value is None or value == 1:
                        return None
                    return lift, value - 1
                if lift == 'continue':
                    continue
                return result
        return None

    def walk__for_stmt(self, node: MapperASTNode) -> MapperASTNode:

        self.scope.add_scope()
        gen = self.walk(node.gen)

        return_value = None

        skips = 0

        for value in gen:
            if skips:
                skips -= 1
                continue
            self.scope.put(self.walk(node.var), value)
            result = self.walk(node.body)
            if result is not None:
                lift, resval = result
                if lift == 'break':
                    if resval is None or resval == 1:
                        break
                    return_value = lift, resval - 1
                    break
                if lift == 'continue':
                    continue

                return_value = result
                return return_value

        self.scope.pop_scope()
        return return_value

    def walk__lift(self, node: MapperASTNode) -> MapperASTNode:
        return (node.lift, self.walk(node.value))

    def walk__amnesia_stmt(self, node: MapperASTNode) -> MapperASTNode:
        self.walk(node.stmt)
        return None

    def walk__call(self, node: MapperASTNode) -> MapperASTNode:

        walked_args = [self.walk(v) for v in node.args]

        function = self.scope.lookup(node.name)

        if isinstance(function, MapperInbuiltFunctionWrapper):
            return function.function(self, walked_args)

        if function is None:
            raise MapperRuntimeError(
                f"function '{node.name}' is not defined")

        if len(function.params) != (len(node.args)):
            raise MapperRuntimeError(
                "Invalid number of positzonal arguments, function " +
                f"'{node.name}' takes {len(function.params)} arguments, " +
                f"but {node.args} were given.")

        scope = self.scope
        newscope = self.scope.copy()
        newscope.add_scope()

        for name, value in zip(function.params, node.args):
            newscope.put_on_last(name, self.walk(value))

        self.scope = newscope
        result = self.walk(function.body)
        self.scope = scope

        if result is None:
            return None

        lift, value = result
        if lift != 'return':
            raise MapperRuntimeError(
                f"A uncaught '{lift}' lift tried to escape function")
        return value

    def walk__neg(self, node: MapperASTNode) -> MapperASTNode:
        if node.op == '!':
            return int(not self.walk(node.val))
        elif node.op == '-':
            return - self.walk(node.val)
        raise MapperRuntimeError(f"Invalid negation operator '{node.op}'")

    def walk__range(self, node: MapperASTNode) -> MapperASTNode:
        start = self.walk(node.start)
        end = self.walk(node.end)
        offset = (1 if end >= start else -1)
        step = self.walk(node.step) if node.step is not None else offset

        if (not isinstance(start, int) or not isinstance(end, int)
                or not isinstance(step, int)):
            raise MapperRuntimeError("Values in range were not of type int")
        return range(start, end + offset, step)

    def walk__assigment(self, node: MapperASTNode) -> MapperASTNode:
        value = self.walk(node.value)
        self.scope.put(node.name, value, min_layer=1)
        return value

    def walk__operation(self, node: MapperASTNode) -> MapperASTNode:
        if node.op not in ARIT_OPS:
            raise MapperRuntimeError(f"Unknown operation '{node.op}")
        left_value = self.walk(node.left)
        right_value = self.walk(node.right)
        if not isinstance(left_value, int) or not isinstance(right_value, int):
            raise MapperRuntimeError("Values in arithmetic operation " +
                                     "were not of type int")
        return ARIT_OPS[node.op](left_value, right_value)

    def walk__variable(self, node: MapperASTNode) -> MapperASTNode:
        value = self.scope.lookup(node.name)
        if value is None:
            raise MapperRuntimeError(f"Variable '{node.name}' is undefined")
        return value

    def walk__integer(self, node: MapperASTNode) -> MapperASTNode:
        return int(node.value)


def in_opt_range(value: int,
                 ran: Optional[Tuple[Optional[int], Optional[int]]]) -> bool:
    if ran is None:
        return True

    left, right = ran

    return (left is None or left <= value) \
        and (right is None or right >= value)


def opt_range_str(rng: Optional[Tuple[Optional[int], Optional[int]]]) \
        -> Tuple[str, bool]:
    if rng is None:
        return 'unlimited number of', True

    bottom, top = rng

    if top is not None and bottom is None:
        return f'at most {top}', top != 1
    if bottom is not None and top is None:
        return f'at least {bottom}', bottom != 1
    if bottom == top:
        return str(bottom), bottom != 1
    return f'from {bottom} to {top}', True


def argument_check(name: str, number: Optional[Tuple[Optional[int],
                                                     Optional[int]]] = None) \
        -> Callable[[MapperInbuiltFunction], MapperInbuiltFunction]:
    def dec(func: MapperInbuiltFunction) -> MapperInbuiltFunction:
        def wrapper(ctx: MapperWalker, args: List[Any]) -> Any:
            if not in_opt_range(len(args), number):
                rng_string, rng_pl = opt_range_str(number)
                raise MapperRuntimeError(
                    f"Function '{name}' takes {rng_string} positional " +
                    f"argument{'s' if rng_pl else ''} " +
                    f"but {len(args)} were given.")
            return func(ctx, args)
        return wrapper
    return dec


@argument_check('print', number=(1, None))
def map_print(mapper: MapperWalker, args: List[Any]) -> None:
    if len(mapper.out) < mapper.max_out:
        mapper.out.append(' '.join(str(a) for a in args))


@argument_check('mov', number=(1, 1))
def mov(mapper: MapperWalker, args: List[Any]) -> None:
    mapper.ship.move_fwd(args[0])


@argument_check('rot', number=(1, 1))
def rot(mapper: MapperWalker, args: List[Any]) -> None:
    mapper.ship.rotate(args[0])


COLS = ['black', 'red', 'green', 'blue']


@argument_check('put', number=(1, 1))
def put(mapper: MapperWalker, args: List[Any]) -> None:
    mapper.ship.put(color=COLS[args[0]])


@argument_check('pen', number=(1, 1))
def pen(mapper: MapperWalker, args: List[Any]) -> None:
    mapper.ship.pendown = bool(args[0])


@argument_check('col', number=(1, 1))
def col(mapper: MapperWalker, args: List[Any]) -> None:
    mapper.ship.color = COLS[args[0]]


INBUILT = {
    'print': map_print,
    'mov': mov,
    'rot': rot,
    'put': put,
    'pen': pen,
    'col': col,
    'BLACK': 0,
    'RED': 1,
    'GREEN': 2,
    'BLUE': 3,
    'RIGHT': 1,
    'LEFT': -1,
    'TRUE': 1,
    'FALSE': 0
}

Model = Any


def compile_code(code: str, parser: Optional[MapperParser] = None) -> Model:
    if parser is None:
        parser = mapper_parser(True)
    processed = mapper_preprocessor(code)
    try:
        return parser.parse(processed)
    except Exception as e:
        raise MapperCompilationError(
            str(e).replace("'{'", 'indent').replace("'}'", 'dedent'))


def compile_codeblocks(codeblocks: List[str],
                       parser: Optional[MapperParser] = None) -> List[Model]:
    return [compile_code(b, parser) for b in codeblocks]


def contains_main(model: Model) -> bool:
    for dec in model['decls']:
        if dec.name == 'main':
            return True
    return False


def run_models(models: List[Model]) -> Tuple[Any, Map, Ship, List[str]]:
    mapper_map = Map(start=(0, 0))
    ship = Ship(mapper_map, mapper_map.start)

    scopestack = ScopeStack()
    scopestack.add_scope()
    walker = MapperWalker(scopestack, INBUILT, ship, mapper_map)
    for model in models:
        for dec in model['decls']:
            walker.walk(dec)
    scopestack.add_scope()
    main_function = scopestack.lookup('main')
    if main_function is None:
        raise MapperRuntimeError("No main found")
    try:
        result = walker.walk(main_function.body)
    except RecursionError:
        raise MapperRuntimeError('maximum recursion depth exceeded')
    except ZeroDivisionError as e:
        raise MapperRuntimeError(str(e))

    return result, mapper_map, ship, walker.out


def generate_level(folder: str, src: str, name: str,
                   prnt_out: bool = False) -> None:

    src = os.path.join(folder, src)
    org_name = name
    name = os.path.join(folder, name)

    with open(f"{src}.mpp") as f:
        code = f.read()

    if name != src:
        with open(f"{name}.mpp", 'w') as f:
            f.write(code)

    models = [compile_code(code)]
    _, mapp, ship, out = run_models(models)
    if prnt_out:
        print(out)

    tiles = export_map(mapp, sep=';')
    with open(f"{name}.txt", "w") as f:
        f.write(tiles)

    start_ship = Ship(mapp)
    render_map(f"{name}.png", mapp, start_ship)
    render_map(f"{name}-res.png", mapp, ship)
    save_json(name, org_name, tiles)


def save_json(path_name: str, org_name: str, test: str) -> None:

    filepath = f"{path_name}.json"
    if os.path.isfile(filepath):
        if False and input('override y/N') != 'y':
            return

    prereqs = []
    desc = f'![{org_name}][{path_name}.png]'

    md_file = f"{path_name}.md"
    if os.path.isfile(md_file):
        with open(md_file) as f:
            header = f.readline()
            prereqs = header[1:].split()
            desc = f.read() + desc
    else:
        with open(md_file, 'w') as f:
            f.write('!\n')

    json_base = f"{path_name}_base.json"
    DocumentEditor().inject({
        'code': org_name,
        'title': org_name,
        'desc': desc,
        'pipeline': 'mapper',
        'tests': [test],
        'prereqs': prereqs
    }).inject_json_file(
        json_base, optional=True
    ).save(filepath)


if __name__ == "__main__":
    name = sys.argv[1] if len(sys.argv) < 3 else sys.argv[2]
    generate_level('mapper', sys.argv[1], name)



from collections import deque
from enum import Enum
import logging
import re
from typing import Callable, Dict, List, Optional, Tuple, Union
import discord
from discord.channel import TextChannel
from discord.errors import InvalidArgument
from discord.ext import commands

from .ellatu import (Ellatu, ImageMessage, Request, TextMessage,
                     MessageSegment, ParagraphMessage, pipeline_sequence)
from .ellatu_db import UserKey

ellatu_logger = logging.getLogger('ellatu')

###############################################################################
# Utils
###############################################################################


def dc_userkey(user: discord.User) -> UserKey:
    return ('dc', str(user.id))


def extract_code_blocks(message: str) -> List[str]:
    result: List[str] = []
    for group in re.findall(r'```[^\n]*([^`]*)```', message):
        result.append(group)
    return result


def starts_in(message: str, character: str) -> bool:
    if len(message) == 0:
        return False
    return message[0] == character


def is_command(message: str) -> bool:
    return starts_in(message.strip(), '!')

###############################################################################
# Message splitting and building
###############################################################################


ParsingType = str
Splitter = Callable[[str], List[Tuple[ParsingType, str]]]


class ParsingNode:
    def __init__(self, p_type: ParsingType, value: Union[str,
                                                         List['ParsingNode']]):
        self.p_type = p_type
        self.value = value

    def split(self, rules: Dict[ParsingType, Splitter]) -> None:
        if isinstance(self.value, str):
            if self.p_type not in rules:
                return
            self.value = [ParsingNode(t, v)
                          for t, v in rules[self.p_type](self.value)]
        for child in self.value:
            child.split(rules)

    def collect(self, res: List[str]) -> List[str]:
        if isinstance(self.value, str):
            res.append(self.value)
        else:
            for child in self.value:
                child.collect(res)
        return res


class ParsingTree:
    def __init__(self, root: Optional[ParsingNode]):
        self.root: Optional[ParsingNode] = root

    def split(self, rules: Dict[ParsingType, Splitter]) -> None:
        if self.root is None:
            return
        self.root.split(rules)

    def collect(self) -> List[str]:
        if self.root is None:
            return []
        return self.root.collect([])


Particle = Tuple[ParsingType, str]


def find_codeblocks(value: str) -> List[Particle]:
    words = value.split('```')
    return [('codeblock', '```' + word + '```')
            if index % 2 == 1 else ('text', word)
            for index, word in enumerate(words)]


def find_codebits(value: str) -> List[Particle]:
    words = value.split('`')
    return [('codebit' if index % 2 == 1 else 'text', word)
            for index, word in enumerate(words)]


def find_lines(value: str) -> List[Particle]:
    lines = value.splitlines()
    return [('atom', w + ('\n' if i < len(lines) - 1 else '')) \
            for i, w in enumerate(lines)]


PARSING_RULES = {
    'paragraph': find_codeblocks,
    'rawtext': find_codebits,
    'text': find_lines
}


def get_atomic_message_parts(message: str) -> List[str]:
    tree = ParsingTree(ParsingNode("paragraph", message))
    tree.split(PARSING_RULES)
    return tree.collect()


class DisMSegment:

    def __init__(self, max_segment_size: int, title: Optional[str] = None,
                 text_blocks: Optional[List[str]] = None):
        self.title = title
        self.text_blocks = text_blocks if text_blocks is not None else []
        self._size = sum([len(t) for t in self.text_blocks])
        self.max_segment_size = max_segment_size
        self.acc = ''

    def add(self, text_block: str, terminator: str = '') -> bool:
        if self.char_size() + len(text_block) > self.max_segment_size:
            return False
        self.text_blocks.append(text_block)
        self._size += len(text_block)
        self.acc += text_block + terminator
        return True

    def char_size(self) -> int:
        return (len(self.title) if self.title else 0) + self._size

    def value(self) -> str:
        return self.acc

    def __bool__(self) -> bool:
        return self.title is not None or bool(self.text_blocks)

    def __str__(self) -> str:
        return str(self.title) + ':\n' + self.value()


class DisMPage:

    def __init__(self, max_page_size: int):
        self.segments: List[DisMSegment] = []
        self._size = 0
        self.max_page_size = max_page_size
        self.images = []

    def add_segment(self, segment: DisMSegment) -> bool:
        if self.char_size() + segment.char_size() > self.max_page_size:
            return False
        self.segments.append(segment)
        self._size += segment.char_size()
        return True

    def char_size(self) -> int:
        return self._size

    def add_image(self, image: Tuple[str, str]):
        self.images.append(image)

    def __bool__(self) -> bool:
        return bool(self.segments) or bool(self.images)

    def __str__(self) -> str:
        return "\n\n".join([str(s) for s in self.segments])


class DisMBuilder:

    def __init__(self, max_segment_size: int = 1000,
                 max_page_size: int = 6000):
        self.pages: List[DisMPage] = []

        self.cur_page: DisMPage = DisMPage(max_page_size)
        self.cur_segment: DisMSegment = DisMSegment(max_segment_size)

        self.max_segment_size = max_segment_size
        self.max_page_size = max_page_size

    def flush_page(self) -> 'DisMBuilder':
        if not self.cur_page:
            return self
        self.pages.append(self.cur_page)
        self.cur_page = DisMPage(self.max_page_size)
        return self

    def flush_segment(self) -> 'DisMBuilder':
        if not self.cur_segment:
            return self
        if not self.cur_page.add_segment(self.cur_segment):
            self.flush_page()
            if not self.cur_page.add_segment(self.cur_segment):
                raise ValueError("Segment doesn't fit into the page")
        self.cur_segment = DisMSegment(self.max_segment_size)
        return self

    def set_title(self, title: str) -> 'DisMBuilder':
        self.cur_segment.title = title
        return self

    def add_text_block(self, text: str) -> 'DisMBuilder':
        if self.cur_segment.add(text.rstrip(), '\n\n'):
            return self
        atoms = get_atomic_message_parts(text)
        for index, atom in enumerate(atoms):
            term, value = ('', atom) \
                if index < len(atoms) - 1 else ('\n\n', atom.rstrip())
            if self.cur_segment.add(value, term):
                continue
            self.flush_segment()
            if not self.cur_segment.add(atom, term):
                raise ValueError("An text atom doesn't fit in empty segment")
        return self

    def add_image(self, image: Tuple[str, str]) -> 'DisMBuilder':
        self.cur_page.add_image(image)
        return self


###############################################################################
# Event
###############################################################################


def create_embed(title: str, color: discord.Color, desc: Optional[str] = None)\
        -> discord.Embed:
    if desc is None:
        return discord.Embed(title=title, color=color)
    return discord.Embed(title=title, color=color, description=desc)


def add_field(embed: discord.Embed, title: Optional[str] = None,
              value: Optional[str] = None, inline: bool = False) -> None:
    embed.add_field(
        name=title if title is not None else '\u200b',
        value=value if value is not None and value.strip() != '' else '-',
        inline=inline
    )


def build_pages(request: Request) -> DisMBuilder:
    builder = DisMBuilder()
    for message in request.messages:
        if isinstance(message, TextMessage):
            builder.add_text_block(message.message)
        elif isinstance(message, ParagraphMessage):
            builder.add_text_block(message.message)
            for image in message.images:
                builder.add_image(image)
        elif isinstance(message, MessageSegment):
            builder.flush_segment().set_title(message.title)
        elif isinstance(message, ImageMessage):
            builder.add_image((message.alt_text, message.location))
        else:
            builder.add_text_block(str(message))
    builder.flush_segment()
    builder.flush_page()
    return builder


async def send_response(request: Request, channel: TextChannel,
                        title: str = "Response",
                        inline: bool = True,
                        desc: Optional[str] = None) -> None:
    color = discord.Color.green() if request.alive else discord.Color.red()
    builder = build_pages(request)

    pages_num = len(builder.pages)
    for index, page in enumerate(builder.pages):
        page_embed = create_embed(title, color,
                                  desc=None if index > 0 else desc)
        for segment in page.segments:
            add_field(page_embed, title=segment.title,
                      value=segment.value(), inline=inline)
        image_file = None
        if page.images:
            _, thumb_file = page.images[0]
            image_file = discord.File(thumb_file, "thumb.png")
            page_embed.set_image(url="attachment://thumb.png")
        if pages_num > 1:
            page_embed.set_footer(text=f"{index + 1}/{pages_num}")
        await channel.send(embed=page_embed, file=image_file)

    request.ellatu.run_request(request.on_resolved(), request)


async def send_error(channel: TextChannel, title: str,
                     message: Optional[str]) -> None:
    embed = discord.Embed(title=f"Oops: {title}")
    if message is not None:
        embed.add_field(name=None, value=message)
    await channel.send(embed=embed)


###############################################################################
# Listening
###############################################################################

class EllatuListeningCog(commands.Cog):
    def __init__(self, bot: commands.Bot, ellatu: Ellatu) -> None:
        self.bot = bot
        self.ellatu = ellatu

    @commands.Cog.listener()
    async def on_ready(self) -> None:
        print(f'{self.bot.user}')

    @commands.Cog.listener()
    async def on_message(self, message) -> None:
        print(f'{message.author}, {message.content}')


###############################################################################
# Commands
###############################################################################

def check_trigger(trigger_event):
    if isinstance(trigger_event, str):
        def pred_str(ctx):
            return hasattr(ctx, 'trigger_event') \
                and ctx.trigger_event == trigger_event
        return pred_str
    elif isinstance(trigger_event, set):
        def pred_set(ctx):
            return hasattr(ctx, 'trigger_event') \
                and ctx.trigger_event in trigger_event
        return pred_set
    raise InvalidArgument("Trigger event can only be str or set")


class EllatuCommandCog(commands.Cog):
    def __init__(self, bot, ellatu):
        self.bot = bot
        self.ellatu = ellatu
        self.event_mode = None

    @commands.command(
        aliases=['hi'],
        help="Sign up for the bot (needs to be done only once, or after a " +
        "name change).",
        brief="sign up"
    )
    @commands.check(check_trigger('on_message'))
    async def hello(self, ctx) -> None:
        self.ellatu.user_connected(dc_userkey(ctx.author), ctx.author.name)
        await ctx.send(f'Hello {ctx.author.name}')

    @commands.command(
        aliases=['lvls'],
        help="Lists levels for current or selected world.",
        brief="lists levels"
    )
    @commands.check(check_trigger('on_message'))
    async def levels(self, ctx, worldcode: Optional[str] = None) -> None:
        request = self.ellatu.run_new_request(
            self.ellatu.get_levels(dc_userkey(ctx.author), worldcode),
            dc_userkey(ctx.author)
        )
        await send_response(request, ctx.channel, title="Levels")

    @commands.command(
        help="Lists all worlds.",
        brief="lists worlds"
    )
    @commands.check(check_trigger('on_message'))
    async def worlds(self, ctx) -> None:
        request = self.ellatu.run_new_request(
            self.ellatu.get_worlds(),
            dc_userkey(ctx.author)
        )
        await send_response(request, ctx.channel, title="Worlds")

    @commands.command(
        aliases=['mov', 'mv'],
        help="Move to a level written as '<worldcode>-<levelcode>' or only " +
        "'<levelcode>' in the current world. Example: " +
        "'!move world-level' or '!move level'.",
        brief="move to level"
    )
    @commands.check(check_trigger('on_message'))
    async def move(self, ctx, levelcode: str) -> None:
        request = self.ellatu.run_new_request(
            self.ellatu.user_move(dc_userkey(ctx.author), levelcode),
            dc_userkey(ctx.author)
        )
        await send_response(request, ctx.channel, title="Move", inline=False)

    @commands.command(
        help="Displays information about the current level.",
        brief="info about level"
    )
    @commands.check(check_trigger('on_message'))
    async def sign(self, ctx) -> None:
        request = self.ellatu.run_new_request(
            self.ellatu.sign_for_user(dc_userkey(ctx.author)),
            dc_userkey(ctx.author)
        )
        await send_response(request, ctx.channel, title="Sign", inline=False)

    @commands.command(
        help="Create workbench with codeblocks, after empty !submit, blocks " +
        "in these codeblocks are submitted. This command wors with " +
        "message edit, meaning you can edit code here and only use " +
        "!submit and !run. (workbenches may not work after time).",
        brief="create workbench"
    )
    async def workbench(self, ctx, *, text: str) -> None:
        codeblocks = extract_code_blocks(text)
        request = self.ellatu.run_new_request(
            self.ellatu.workbench(dc_userkey(ctx.author), codeblocks),
            dc_userkey(ctx.author)
        )
        if request.alive:
            await ctx.send(f"Workbench saved for {ctx.author.name}")
        else:
            await send_response(request, ctx.channel, title="Workbench")

    @commands.command(
        aliases=['s'],
        help="Submit codeblocks (blocks created using triple `), these " +
        "codeblocks are only saved, they still need to be run. If there " +
        "are no blocks in the message, blocks from !workbench are used.",
        brief="submit codeblocks"
    )
    @commands.check(check_trigger('on_message'))
    async def submit(self, ctx, *, text: Optional[str] = None) -> None:
        codeblocks = extract_code_blocks(text) if text is not None else None
        request = self.ellatu.run_new_request(
            self.ellatu.submit(dc_userkey(ctx.author), codeblocks),
            dc_userkey(ctx.author)
        )
        await send_response(request, ctx.channel, title="Submit")

    @commands.command(
        aliases=['r'],
        help="Run submitted codeblocks against tests from current level, " +
        "tag other users to submit their codeblocks. (This feature may " +
        "or may not be allowed, depending on the level.",
        brief="run submitted blocks"
    )
    @commands.check(check_trigger('on_message'))
    async def run(self, ctx, users: commands.Greedy[discord.Member]) -> None:
        userkeys = [dc_userkey(u) for u in reversed(users)] + \
            [dc_userkey(ctx.author)]
        request = self.ellatu.run_new_request(
            self.ellatu.run(userkeys),
            dc_userkey(ctx.author)
        )
        await send_response(request, ctx.channel, title="Run")

    @commands.command(
        help="Submit and run combined, the fact that run failed doesn't mean" +
        " that submit failed also (no rollback).",
        brief="submit and run"
    )
    @commands.check(check_trigger('on_message'))
    async def subrun(self, ctx, users: commands.Greedy[discord.Member], *,
                     text: Optional[str] = None) -> None:
        codeblocks = extract_code_blocks(text) if text is not None else None
        userkeys = [dc_userkey(ctx.message.author)] + \
            [dc_userkey(u) for u in users]
        request = self.ellatu.run_new_request(
            pipeline_sequence([
                self.ellatu.submit(dc_userkey(ctx.author), codeblocks),
                self.ellatu.run(userkeys)
            ]),
            dc_userkey(ctx.author)
        )
        await send_response(request, ctx.channel, title="Submit and Run")

    @commands.command(
        help="Draw map of the current or selected world.",
        brief="draw map"
    )
    @commands.check(check_trigger('on_message'))
    async def map(self, ctx, worldcode=None) -> None:
        request = self.ellatu.run_new_request(
            self.ellatu.draw_map(dc_userkey(ctx.message.author), worldcode),
            dc_userkey(ctx.author)
        )
        await send_response(request, ctx.channel, title="Map")

    # Listeners ###

    async def process_commands(self, message, trigger_event) -> None:
        if message.author.bot:
            return
        ctx = await self.bot.get_context(message)
        ctx.trigger_event = trigger_event
        await self.bot.invoke(ctx)

    @commands.Cog.listener()
    async def on_message(self, message) -> None:
        await self.process_commands(message, "on_message")

    @commands.Cog.listener()
    async def on_message_edit(self, _, message) -> None:
        await self.process_commands(message, "on_message_edit")

    @commands.Cog.listener()
    async def on_command_error(self, ctx, error):
        if isinstance(error,
                      (commands.CommandNotFound,
                       commands.MissingPermissions)):
            await ctx.send(f"{str(error)}")
        else:
            await ctx.send(f"Oops, internal error occured, contact admin")
            ellatu_logger.log(logging.ERROR, str(error))



import json
import os
from typing import Optional, Dict, List
from .ellatu_db import Document

class DocumentEditor:

    def __init__(self, json: Optional[Document] = None):
        self._doc = json if json is not None else {}

    def inject(self, doc: Document) -> 'DocumentEditor':
        for key, value in doc.items():
            self._doc[key] = value
        return self

    def select(self, keys: List[str]) -> 'DocumentEditor':
        new_doc = {}
        for key in keys:
            if key in self._doc:
                new_doc[key] = self._doc[key]
        self._doc = new_doc
        return self

    def relabel(self, relabling: Dict[str, str]) -> 'DocumentEditor':
        for oldkey, newkey in relabling.items():
            self._doc[newkey] = self._doc[oldkey]
            del self._doc[oldkey]
        return self

    def inject_json_file(self, filename: str,
                         optional: bool = False) -> 'DocumentEditor':
        if not optional or os.path.isfile(filename):
            with open(filename, 'r') as f:
                json_obj = json.load(f)
                for key, value in json_obj.items():
                    self._doc[key] = value
        return self

    def save(self, filename: str) -> 'DocumentEditor':
        with open(filename, 'w') as f:
            json.dump(self._doc, fp=f)
        return self

    def print(self, *args, **kwargs) -> 'DocumentEditor':
        print(json.dumps(self._doc, *args, **kwargs))
        return self

    def mat(self) -> Document:
        return self._doc



from collections import deque
import os
import re

from typing import (Deque, Dict, List, Callable,
                    Optional, Any, Set, Tuple, TypeVar)
from enum import Enum
from datetime import datetime, timedelta
from random import randint
import logging
import pygraphviz

from . import image_editing as imge
from .ellatu_db import (Document, EllatuDB, MongoId,
                        UserKey, LevelKey, get_levelkey, get_userkey)

###############################################################################
# Logger
###############################################################################

logger = logging.getLogger('ellatu')

###############################################################################
# Types
###############################################################################

F = TypeVar('F', bound=Callable[..., Any])

###############################################################################
# Misc.
###############################################################################


def level_code_parts(worldcode: str, levelcode: str) -> str:
    return f"{worldcode}-{levelcode}"


def level_code_doc(level: Document) -> str:
    return level_code_parts(level["worldcode"], level["code"])


def level_code_key(levelkey: LevelKey) -> str:
    return level_code_parts(levelkey[0], levelkey[1])

###############################################################################
# Messages
###############################################################################


class MessageType(Enum):
    LOG = 1
    TRACE = 2
    ERROR = 3


class Message:

    def __init__(self, message_type: MessageType):
        self.message_type = message_type

    def prefix(self, message_body: str) -> str:
        return f"[{self.message_type}]: {message_body}"


class MessageSegment(Message):

    def __init__(self, title: str,
                 message_type: MessageType = MessageType.LOG):
        super().__init__(message_type)
        self.title = title

    def __str__(self) -> str:
        return self.prefix(self.title)


class TextMessage(Message):

    def __init__(self, message: str,
                 message_type: MessageType = MessageType.LOG):
        super().__init__(message_type)
        self.message = message

    def __str__(self) -> str:
        return self.prefix(self.message)


class ParagraphMessage(Message):

    def __init__(self, message: str,
                 message_type: MessageType = MessageType.LOG):
        super().__init__(message_type)

        matches = re.split(r'!(\[[^\]]*\]\[[^\]]*\])',
                           message)
        images = []
        final_message = ""
        for index, part in enumerate(matches):
            if index % 2 == 0:
                final_message += part
            else:
                worlds = part[1:-1].split('][', maxsplit=1)
                images.append(worlds)
                final_message += f"_[{len(images)}]_"

        self.message = final_message
        self.images = images

    def __str__(self) -> str:
        return self.prefix(self.message)


class ImageMessage(Message):
    def __init__(self, alt_text: str, location: str,
                 message_type: MessageType = MessageType.LOG) -> None:
        super().__init__(message_type)
        self.alt_text = alt_text
        self.location = location

    def __str__(self) -> str:
        return self.prefix(f"Image [{self.location}][{self.alt_text}]")


###############################################################################
# Request
###############################################################################


class Request:

    def __init__(self, ellatu: "Ellatu", requestor: UserKey):

        self.ellatu = ellatu
        self.id = ellatu.get_req_id()

        self.alive = True
        self.messages: List[Message] = []

        self.level: Optional[Document] = None
        self.levels: Dict[LevelKey, Document] = {}

        self.requestor: UserKey = requestor
        self.users: Dict[UserKey, Document] = {}

        self.codeblocks: Dict[MongoId, List[str]] = {}

        self.data: Dict[str, Any] = {}
        self.temp_files: Dict[str, str] = {}

        self._on_resolved_actions: List['RequestAction'] = []

    def add_on_res(self, action: 'RequestAction') -> None:
        self._on_resolved_actions.append(action)

    def on_resolved(self) -> 'RequestAction':
        return pipeline_sequence(self._on_resolved_actions)

    def add_message(self, message: Message) -> None:
        self.messages.append(message)

    def __str__(self) -> str:
        res = f"Request: [{'Alive' if self.alive else 'Dead'}]\n"
        for message in self.messages:
            res += str(message) + '\n'
        return res


RequestAction = Callable[[Request], Request]


# Message Utils ###############################################################


def log_msg(request: Request, message: str,
            message_type: MessageType = MessageType.LOG) -> Request:
    request.add_message(TextMessage(message, message_type=message_type))
    return request


def trace(request: Request, message: str) -> Request:
    return log_msg(request, message, MessageType.TRACE)


def log(request: Request, message: str) -> Request:
    return log_msg(request, message, MessageType.LOG)


def error(request: Request, message: str) -> Request:
    return log_msg(request, message, MessageType.ERROR)


def add_msg(message: Message) -> RequestAction:
    def action(request: Request) -> Request:
        request.add_message(message)
        return request
    return action


def terminate_request(request: Request, message: str) -> Request:
    request.alive = False
    return error(request, message)


def const_action(request: Request) -> Request:
    return request


def kill_request(request: Request) -> Request:
    return terminate_request(request, "The request has been killed")


# Action utils ################################################################

ExtRequestAction = Callable[..., Request]


def data_action(keys: List[str]) \
        -> Callable[[ExtRequestAction], RequestAction]:
    def wrapper(func: ExtRequestAction) -> RequestAction:
        def action(request: Request) -> Request:
            values = []
            for key in keys:
                if key not in request.data:
                    return terminate_request(request, "Internal error")
                values.append(request.data[key])
            return func(request, *values)
        return action
    return wrapper


def pipeline_sequence(actions: List[RequestAction]) -> RequestAction:

    def action(request: Request) -> Request:
        for action in actions:
            request = action(request)
            if not request.alive:
                break
        return request

    return action


def pipeline_tree(tree: Dict[str, RequestAction]) -> RequestAction:
    def pipeline_action(request: Request) -> Request:
        if request.level is None:
            return terminate_request(request, "No level set")
        if request.level['pipeline'] not in tree:
            return terminate_request(request, "Unknown pipeline")
        return tree[request.level['pipeline']](request)
    return pipeline_action


def pipeline_with(action: Callable[[Request], RequestAction]) -> RequestAction:
    def with_action(request: Request) -> Request:
        return action(request)(request)
    return with_action

# File in action ##############################################################


def set_data(key: str, value: Any) -> RequestAction:
    def set_data_action(request: Request) -> Request:
        request.data[key] = value
        return request
    return set_data_action


def add_on_resolved(action: RequestAction) -> RequestAction:
    def add_on_resolved_action(request: Request) -> Request:
        request.add_on_res(action)
        return request
    return add_on_resolved_action


def add_temp_file(extension: str) -> RequestAction:
    def create_temp_file_action(request: Request) -> Request:
        filename = request.ellatu.temp_files.add_temp_filename(
            str(request.id) + '-' + extension
        )
        request.temp_files[extension] = filename
        return request
    return create_temp_file_action


def clean_temp_files() -> RequestAction:
    def clean_temp_files_action(request: Request) -> Request:
        for _, filename in request.temp_files.items():
            request.ellatu.temp_files.remove_temp_file(filename)
        request.temp_files.clear()
        return request
    return clean_temp_files_action


def remove_files(filenames: List[str]) -> RequestAction:
    def action(request: Request) -> Request:
        for filename in filenames:
            os.remove(filename)
        return request
    return action

# User actions ################################################################


def add_users(userkeys: List[UserKey]) -> RequestAction:
    def action(request: Request) -> Request:
        keys_set = set(userkeys)
        users = request.ellatu.db.user.get_users(userkeys)
        for user in users:
            userkey = get_userkey(user)
            if userkey in keys_set:
                keys_set.remove(userkey)
                request.users[get_userkey(user)] = user
        if keys_set:
            terminate_request(request,
                              "The users were not found")
        return request
    return action


def move_users() -> RequestAction:
    def action(request: Request) -> Request:
        if request.level is None:
            return terminate_request(request, "Invalid level")
        for userkey, _ in request.users.items():
            if not request.ellatu.db.user.move_user(
                    userkey, request.level["worldcode"],
                    request.level["code"]):
                return terminate_request(request, "Unable to move user")
        return request
    return action

# Codeblocks actions ##########################################################


def assign_codeblocks(assignments: Dict[UserKey, List[str]]) -> RequestAction:
    def assign_codeblocks_action(request: Request) -> Request:
        for userkey, codeblocks in assignments.items():
            if userkey not in request.users:
                request.messages.append(TextMessage(
                    "User not is not in the request"))
                continue
            user = request.users[userkey]
            request.codeblocks[user["_id"]] = codeblocks
        return request
    return assign_codeblocks_action


def assign_from_workplace(userkeys: List[UserKey]) -> RequestAction:
    def assign_from_workplace_action(request: Request) -> Request:
        if request.level is None:
            return terminate_request(request,
                                     "The request doesn't have valid \
                                        level set")
        users = request.ellatu.db.user.get_users(userkeys)
        userids = [u["_id"] for u in users]
        codeblocks = request.ellatu.db.workplace\
            .get_codeblocks(userids, request.level["worldcode"])

        for userid, blocks in codeblocks.items():
            request.codeblocks[userid] = blocks
        return request
    return assign_from_workplace_action


def assign_from_workbench() -> RequestAction:
    def assign_from_workbench_action(request: Request) -> Request:
        if request.level is None:
            return terminate_request(request, "Level not set")
        userids = [u["_id"] for _, u in request.users.items()]
        codeblocks = request.ellatu.db.workplace\
            .get_workbenches(userids, request.level["worldcode"])
        for _, user in request.users.items():
            userid = user["_id"]
            if userid in codeblocks:
                request.codeblocks[userid] = codeblocks[userid]
        return request
    return assign_from_workbench_action


def assign_to_workbench() -> RequestAction:
    def assign_to_workbench_action(request: Request) -> Request:
        if request.level is None:
            return terminate_request(request, "Level not set")
        if not request.ellatu.db.workplace.set_workbenches(
                request.codeblocks, request.level["worldcode"]):
            return terminate_request(request, "Failed to save workbench")
        return request
    return assign_to_workbench_action


def print_codeblocks() -> RequestAction:
    def print_codeblocks_action(request: Request) -> Request:
        res = ""

        usermap = {}
        for userkey, user in request.users.items():
            usermap[user["_id"]] = userkey

        for userid, codeblocks in request.codeblocks.items():
            if userid not in usermap:
                return terminate_request(request, "Users not loaded corretly")

            user = request.users[usermap[userid]]
            res += user["username"]
            for codeblock in codeblocks:
                res += "```\n" + codeblock + "```"

        trace(request, res)

        return request
    return print_codeblocks_action


def limit_users(number: Optional[int]) -> RequestAction:
    def limit_users_action(request: Request) -> Request:
        if number is None:
            return request
        if len(request.codeblocks.keys()) > number:
            return terminate_request(
                request,
                f"There were more than {number} users")
        return request
    return limit_users_action


def limit_codeblocks(number: Optional[int]) -> RequestAction:
    def limit_codeblocks_action(request: Request) -> Request:
        if number is None:
            return request
        for _, codeblocks in request.codeblocks.items():
            if len(codeblocks) > number:
                return terminate_request(
                    request,
                    f"There were more then {number} codeblocks"
                )
        return request
    return limit_codeblocks_action


def limit_lines(number: Optional[int]) -> RequestAction:
    def limit_lines_action(request: Request) -> Request:
        if number is None:
            return request
        for _, codeblocks in request.codeblocks.items():
            for codeblock in codeblocks:
                if len(list(filter(lambda x: x.strip() != '',
                                   codeblock.splitlines()))) > number:
                    return terminate_request(
                        request,
                        f"A codeblock was longer than {number} lines"
                    )
        return request
    return limit_lines_action


def limit_columns(number: Optional[int]) -> RequestAction:
    def limit_columns_action(request: Request) -> Request:
        if number is None:
            return request
        for _, codeblocks in request.codeblocks.items():
            for codeblock in codeblocks:
                for line in codeblock.splitlines():
                    if len(line.rstrip()) > number:
                        return terminate_request(
                            request,
                            f"Line was longer than {number} characters"
                        )
        return request
    return limit_columns_action


# Localize actions ############################################################


def parse_level_code(code: str) -> Tuple[Optional[str], str]:
    match = re.match(r'^(\w+)\-(\w+)$', code)
    if match:
        return match.group(1), match.group(2)
    return None, code


def localize_by_user(userkey: UserKey) -> RequestAction:
    def localize_by_user_action(request: Request) -> Request:
        user = request.ellatu.db.user.get_user(userkey)
        if user is None:
            return terminate_request(request, "The user was not found")
        worldcode = user['worldcode']
        levelcode = user['levelcode']
        if worldcode is None or levelcode is None:
            return terminate_request(request, "The user is in no level")
        request.level = request.ellatu.db.level.get_by_code(worldcode,
                                                            levelcode)
        return request
    return localize_by_user_action


def localize_by_code(worldcode: Optional[str], levelcode: Optional[str],
                     userkey: Optional[UserKey] = None) -> RequestAction:
    def localize_by_code_action(request: Request) -> Request:
        w_code = worldcode
        l_code = levelcode
        if w_code is None or l_code is None:
            if userkey is None:
                return terminate_request(
                    request, "Part of code missing but no user")
            user = request.ellatu.db.user.get_user(userkey)
            if user is None:
                return terminate_request(
                    request, "Part of code missing but user not found")
            if user['worldcode'] is None or user['levelcode'] is None:
                return terminate_request(request, "The user is in no level")
            if w_code is None:
                w_code = user['worldcode']
            if l_code is None:
                l_code = user['levelcode']

        level = request.ellatu.db.level.get_one(
            worldcode=w_code, code=l_code)
        if not level:
            return terminate_request(request, "Invalid level code")
        request.level = level
        return request
    return localize_by_code_action


# Level/world actions #########################################################


def add_levels_worldcode(worldcode: str) -> RequestAction:
    def add_levels_worldcode_action(request: Request) -> Request:
        levels = request.ellatu.db.level.get(worldcode=worldcode)
        for level in levels:
            request.levels[get_levelkey(level)] = level
        return request
    return add_levels_worldcode_action


def add_local_levels() -> RequestAction:
    def add_local_levels_action(request: Request) -> Request:
        if request.level is None:
            return terminate_request(request, "Level not set")
        return add_levels_worldcode(request.level['worldcode'])(request)
    return add_local_levels_action


def print_levels() -> RequestAction:
    @data_action(["beaten"])
    def print_levels_action(request: Request,
                            beaten: Set[Tuple[str, str]]) -> Request:
        res = ""
        for _, level in request.levels.items():
            title_str = f"**{level['title']}** [*{level_code_doc(level)}*]"

            locked = is_locked(beaten, level)
            done = is_beaten(beaten, level)
            title_wrapper = ('', '')
            if done:
                title_wrapper = ('__', '__')
            elif locked:
                title_wrapper = ('|| Locked: ', '||')
            res += title_wrapper[0] + title_str + title_wrapper[1]
            if level['prereqs']:
                p_str = ','.join(['_' + s + '_' for s in level['prereqs']])
                res += f" needs {p_str}"
            res += '\n'
        return trace(request, res)
    return print_levels_action


def draw_levels(filekey: str, userkey: Optional[UserKey] = None,
                include_worldcode: bool = True) -> RequestAction:
    @data_action(["beaten"])
    def draw_levels_action(request: Request,
                           beaten: Set[Tuple[str, str]]) -> Request:

        if filekey not in request.temp_files:
            return terminate_request(request, "Temp file wasn't ready")
        filename = request.temp_files[filekey]

        dot = pygraphviz.AGraph(strict=False, directed=True)
        dot.node_attr["shape"] = "rect"
        dot.node_attr["style"] = "filled"
        target: Optional[Tuple[str, str]] = None

        if userkey is not None:
            if userkey not in request.users:
                return terminate_request(request, "User not loaded")
            user = request.users[userkey]
            target = (user['worldcode'], user['levelcode'])

        for _, level in request.levels.items():
            fillcolor = "white"
            if is_beaten(beaten, level):
                fillcolor = "#56de3e"
            if is_locked(beaten, level):
                fillcolor = "#d63c31"
            if target and target == (level['worldcode'],
                                     level['code']):
                fillcolor = "#5c70d6"
            code = level_code_doc(level)
            label = code if include_worldcode else level['code']
            dot.add_node(code,  fillcolor=fillcolor, label=label)
            for prereq in level['prereqs']:
                dot.add_edge(level_code_parts(
                    level['worldcode'], prereq), code)
        dot.layout(prog='dot')
        dot.draw(filename)
        imge.edit_image(filename, imge.edit_sequence([
            imge.expand_to_aspect(1),
            imge.expand(total=20)
        ]))
        return request
    return draw_levels_action


def print_level_info(
        header: Callable[[Document], Optional[str]] = lambda _: None,
        desc: bool = True) -> RequestAction:
    def print_level_info_action(request: Request) -> Request:
        if request.level is None:
            return trace(request, "No level selected")
        text = f"**{request.level['title']}** " + \
            f"[_{level_code_doc(request.level)}_]"

        if desc:
            header_text = header(request.level)
            if header_text:
                text += '\n' + header_text
            text += '\n' + request.level['desc']

        request.add_message(ParagraphMessage(text))
        return request
    return print_level_info_action


def format_world_title(world: Document) -> str:
    return f"**{world['title']}** [_{world['code']}_]\n"


def print_worlds() -> RequestAction:
    def print_worlds_action(request: Request) -> Request:
        worlds = request.ellatu.db.world.get()
        res = ""
        for world in worlds:
            res += format_world_title(world)
        return trace(request, res)
    return print_worlds_action


def print_world_info(worldcode: str) -> RequestAction:
    def print_world_info_action(request: Request) -> Request:
        world = request.ellatu.db.world.get_one(code=worldcode)
        if world is None:
            return terminate_request(request, "World not found")
        return trace(request, format_world_title(world))
    return print_world_info_action


def print_world_info_local() -> RequestAction:
    def print_world_info_local_action(request: Request) -> Request:
        if request.level is None:
            return terminate_request(request, "Level not set")
        return print_world_info(request.level['worldcode'])(request)
    return print_world_info_local_action


def print_world_info_user(userkey: UserKey) -> RequestAction:
    def print_world_info_user_action(request: Request) -> Request:
        if userkey not in request.users:
            return terminate_request(request, "User not in request")
        user = request.users[userkey]
        if not user['worldcode']:
            return terminate_request(request, "User not in world")
        return print_world_info(user['worldcode'])(request)
    return print_world_info_user_action

# Permission actions ##########################################################


def beaten_from_solutions(solutions: List[Document]) -> Set[Tuple[str, str]]:
    return set([(s['worldcode'], s['levelcode']) for s in solutions])


def is_locked(beaten: Optional[Set[Tuple[str, str]]], level: Document) -> bool:
    if beaten is None:
        return False
    for pre in level['prereqs']:
        if (level['worldcode'], pre) not in beaten:
            return True
    return False


def is_beaten(beaten: Optional[Set[Tuple[str, str]]], level: Document) -> bool:
    if beaten is None:
        return False
    return (level['worldcode'], level['code']) in beaten


def permission_check() -> RequestAction:
    def permission_check_action(request: Request) -> Request:
        if request.level is None:
            return trace(request, "No level to check permission for")

        not_allowed_users = []
        for _, user in request.users.items():
            sols = request.ellatu.db.solution.get_solutions(
                user["_id"], request.level['worldcode'])
            beaten = beaten_from_solutions(sols)
            if is_locked(beaten, request.level):
                not_allowed_users.append(user['username'])
        if not_allowed_users:
            return terminate_request(
                request,
                "Users are not allowed to be in the level: " +
                f"{','.join(not_allowed_users)}")
        return request
    return permission_check_action


# Submission actions ##########################################################


def save_submit() -> RequestAction:
    def save_submit_action(request: Request) -> Request:

        if request.level is None:
            return terminate_request(request, "Invalid level")

        for userid, codeblocks in request.codeblocks.items():
            if not request.ellatu.db.workplace.add_submission(
                userid,
                request.level["worldcode"],
                codeblocks
            ):
                terminate_request(request, "Unable to save submit")
                return request
        return request
    return save_submit_action


def save_solution() -> RequestAction:
    def save_solution_action(request: Request) -> Request:
        if not request.alive:
            return terminate_request(request, "The request is already dead")
        if request.level is None:
            return terminate_request(request, "Invalid level")
        for userid, _ in request.codeblocks.items():
            if not request.ellatu.db.solution.add_solution(
                userid,
                request.level['worldcode'],
                request.level['code'],
                3
            ):
                return terminate_request(request, "Unable to save solution")

        return trace(request, "The scores have been saved")
    return save_solution_action


def load_beaten_by_user(userkey: UserKey) -> RequestAction:
    def load_beaten_by_user_action(request: Request) -> Request:
        user = request.ellatu.db.user.get_user(userkey)
        if user is None:
            return terminate_request(request, "User not found")
        solutions = request.ellatu.db.solution.get_solutions(user["_id"])
        beaten = set([(s['worldcode'], s['levelcode']) for s in solutions])
        request.data["beaten"] = beaten
        return request
    return load_beaten_by_user_action


###############################################################################
# Ellatu interface
###############################################################################

class TempFileStorage:

    def __init__(self, temp_folder: str = 'ellatu_temp'):
        self.temp_folder = temp_folder
        self.temp_files: Deque[Tuple[datetime, str]] = deque()

    def add_temp_filename(self, name: str) -> str:
        filename = os.path.join(self.temp_folder, name)
        self.temp_files.append((datetime.now(), filename))
        return filename

    def remove_temp_file(self, filename: str) -> None:
        os.remove(filename)

    def remove_temp_files(self) -> None:
        if not self.temp_files:
            return
        dumpdate = datetime.now() - timedelta(hours=1)
        while self.temp_files[0][0] < dumpdate:
            _, temp_file = self.temp_files.popleft()
            os.remove(temp_file)


class Ellatu:

    def __init__(self, ellatu_db: EllatuDB, temp_folder: str = 'ellatu_temp'):
        self.on_submit_workflow: RequestAction = const_action
        self.on_run_workflow: RequestAction = const_action
        self.header: Callable[[Document], Optional[str]] = lambda _x: None
        self.db = ellatu_db
        self.temp_files = TempFileStorage(temp_folder=temp_folder)

    def run_request(self, request_action: RequestAction,
                    request: Request) -> Request:
        try:
            return request_action(request)
        except Exception as error:
            logger.exception(error)
            return terminate_request(request, "Unexpected internal error")

    def run_new_request(self, request_action: RequestAction,
                        requestor: UserKey) -> Request:
        logger.info(f"{requestor} made an request")
        request = Request(self, requestor)
        return self.run_request(request_action, request)

    def user_move(self, userkey: UserKey, code: str) -> RequestAction:
        worldcode, levelcode = parse_level_code(code)
        return pipeline_sequence([
            add_users([userkey]),
            localize_by_code(worldcode, levelcode, userkey),
            permission_check(),
            move_users(),
            add_msg(TextMessage("**You have been moved to:**")),
            print_level_info(header=self.header)
        ])

    def workbench(self, userkey: UserKey,
                  codeblocks: List[str]) -> RequestAction:
        return pipeline_sequence([
            add_users([userkey]),
            localize_by_user(userkey),
            permission_check(),
            assign_codeblocks({userkey: codeblocks}),
            assign_to_workbench()
        ])

    def submit(self, userkey: UserKey,
               codeblocks: Optional[List[str]]) -> RequestAction:
        return pipeline_sequence([
            add_users([userkey]),
            localize_by_user(userkey),
            permission_check(),

            assign_codeblocks({userkey: codeblocks})
            if codeblocks is not None
            else assign_from_workbench(),

            self.on_submit_workflow,
            save_submit(),
            add_msg(MessageSegment("The codeblocks were added to:")),
            print_level_info(header=self.header, desc=False)
        ])

    def run(self, userkeys: List[UserKey]) -> RequestAction:
        return pipeline_sequence([
            add_users(userkeys),
            localize_by_user(userkeys[0]),
            permission_check(),
            assign_from_workplace(userkeys),
            add_msg(MessageSegment("Running following blocks:")),
            print_codeblocks(),
            set_data('blocks_order', userkeys),
            self.on_run_workflow,
            save_solution()
        ])

    def get_worlds(self) -> RequestAction:
        return pipeline_sequence([
            add_msg(MessageSegment('Available worlds')),
            print_worlds()
        ])

    def get_levels(self, userkey: UserKey,
                   worldcode: Optional[str] = None) -> RequestAction:
        return pipeline_sequence([
            pipeline_sequence(
                [localize_by_user(userkey), add_local_levels()]
            ) if worldcode is None else add_levels_worldcode(worldcode),
            load_beaten_by_user(userkey),
            add_msg(MessageSegment('Available levels in _mapper_')),
            print_levels()
        ])

    def draw_map(self, userkey: UserKey,
                 worldcode: Optional[str] = None) -> RequestAction:
        return pipeline_sequence([
            add_users([userkey]),
            pipeline_sequence(
                [localize_by_user(userkey), add_local_levels()]
            ) if worldcode is None else add_levels_worldcode(worldcode),
            load_beaten_by_user(userkey),
            print_world_info_user(
                userkey) if worldcode is None else print_world_info(worldcode),
            add_temp_file('map.png'),
            add_on_resolved(clean_temp_files()),
            draw_levels('map.png', userkey=userkey, include_worldcode=False),
            pipeline_with(lambda request: pipeline_sequence([
                add_msg(ImageMessage('map', request.temp_files['map.png']))
            ]))
        ])

    def sign_for_user(self, userkey: UserKey) -> RequestAction:
        return pipeline_sequence([
            add_users([userkey]),
            localize_by_user(userkey),
            permission_check(),
            print_level_info(header=self.header)
        ])

    def user_connected(self, userkey: UserKey, username: str) -> bool:
        return self.db.user.open_user(userkey, username) is not None

    def get_req_id(self) -> int:
        return randint(0, 1000000000000)


class EllatuPipeline:

    def on_submit(self) -> RequestAction:
        return const_action

    def on_run(self) -> RequestAction:
        return const_action



from pprint import pprint
from typing import Any, Dict, Generic, List, Optional, Set, Tuple, TypeVar
from pymongo import MongoClient
from datetime import datetime

from pymongo.collection import ReturnDocument

DB_DEV = "ellatu_dev"

COL_USERS = "users"
COL_LEVELS = "levels"
COL_WORKPLACES = "workplaces"
COL_SOLUTIONS = "solutions"
COL_WORLDS = "worlds"

MAX_CODEBLOCKS = 3

Value = Any

T = TypeVar("T")

Collection = Any
Document = Dict[Any, Any]
MongoId = str

ClientService = str
ClientId = str
UserKey = Tuple[ClientService, ClientId]
Worldcode = str
Levelcode = str
LevelKey = Tuple[Worldcode, Levelcode]


###############################################################################
# Misc
###############################################################################

def int_in_range(value: int, min_value: Optional[int],
                 max_value: Optional[int]) -> bool:
    if min_value is not None and value < min_value:
        return False
    if max_value is not None and value > max_value:
        return False
    return True


def get_userkey(doc: Document) -> UserKey:
    return (doc["client_ser"], doc['client_id'])


def get_levelkey(level: Document) -> LevelKey:
    return (level["worldcode"], level["code"])

###############################################################################
# Validation
###############################################################################


def val_error(trace: List[str], message: str) -> bool:
    trace.append(message)
    return False


class Validator(Generic[T]):

    def __init__(self, message: str = "invalid value"):
        self.message = message

    def pred(self, value: T, trace: List[str]) -> bool:
        return False

    def validate(self, value: T) -> bool:
        trace: List[str] = []
        if not self.pred(value, trace):
            print("Following value is invalid: ")
            pprint(value)
            print("\n".join(trace))
            return False
        return True


class StringValidator(Validator[str]):

    def __init__(self, min_size: Optional[int] = None,
                 max_size: Optional[int] = None,
                 charset: Optional[Set[str]] = None):
        super().__init__(message="invalid string")
        self.min_size = min_size
        self.max_size = max_size
        self.charset = charset

    def pred(self, value: str, trace: List[str]) -> bool:
        if not int_in_range(len(value), self.min_size, self.max_size):
            return val_error(trace, "Invalid string length")
        if self.charset:
            for c in value:
                if c not in self.charset:
                    return val_error(trace, "Character not in the allowed " +
                                     "charset")
        return True


class IntegerValidator(Validator[int]):
    def __init__(self, min_value: Optional[int] = None,
                 max_value: Optional[int] = None):
        super().__init__(message="invalid number")
        self.min_value = min_value
        self.max_value = max_value

    def pred(self, value: int, trace: List[str]) -> bool:
        if not isinstance(value, int):
            return val_error(trace, "Not integer")
        if not int_in_range(value, self.min_value, self.max_value):
            return val_error(trace, f"Not in the range {self.min_value}"
                             + f" {self.max_value}")
        return True


class RefKeyValidator(Validator[T]):

    def __init__(self, collection: Collection, atr_name: str):
        super().__init__("key not present in collection")
        self.collection = collection
        self.atr_name = atr_name

    def pred(self, value: T, trace: List[str]) -> bool:
        doc = {}
        doc[self.atr_name] = value
        if not self.collection.find_one(doc):
            return val_error(trace, f"Reference ({value}) not in the"
                             + " collection")
        return True


class RefValidator(Validator[Document]):

    def __init__(self, collection: Collection, keys: Dict[str, str]):
        super().__init__("value not present")
        self.collection = collection
        self.keys = keys

    def pred(self, value: Document, trace: List[str]) -> bool:
        mask = {}
        for key in self.keys:
            if key not in value:
                return val_error(trace, "The key referenced is not in the"
                                 + " value")
            mask[self.keys[key]] = value[key]
        if not self.collection.find_one(mask):
            return val_error(trace, "The reference is not present")
        return True


class PrimaryKeyValidator(Validator[Document]):

    def __init__(self, collection: Collection, keys: List[str]):
        super().__init__("primary key already present")
        self.collection = collection
        self.keys = keys

    def pred(self, value: Document, trace: List[str]) -> bool:
        mask = {}
        for key in self.keys:
            if key not in value:
                return val_error(trace, "The key that is part of the primary"
                                 + "key is not present")
            mask[key] = value[key]
        if self.collection.find_one(mask):
            return val_error(trace, "The primary key is already present")

        return True


class ReqValidator(Validator[Optional[Any]]):

    def __init__(self) -> None:
        super().__init__("value is required")

    def pred(self, value: Optional[Any], _: List[str]) -> bool:
        return value is not None


class ReqFieldsValidator(Validator[Document]):
    def __init__(self, keys: List[str]):
        super().__init__("required value is not present")
        self.keys = keys

    def pred(self, value: Document, trace: List[str]) -> bool:
        for key in self.keys:
            if key not in value or value[key] is None:
                return val_error(trace, f"The required value ({key}) is " +
                                 "not present")
        return True


class DictValidator(Validator[Document]):

    def __init__(self, scheme: Dict[str, Validator[Any]]):
        super().__init__("value not in scheme")
        self.scheme = scheme

    def pred(self, value: Document, trace: List[str]) -> bool:

        if not isinstance(value, dict):
            return val_error(trace, "The value is not dictionary")

        for key in self.scheme:
            if key not in value:
                return val_error(trace, f"Key {key} is not present")
            if not self.scheme[key].pred(value[key], trace):
                return val_error(trace, f"The validator failed on key: {key}")

        return True


class SequentialValidator(Validator[T]):

    def __init__(self, validators: List[Validator[T]]):
        super().__init__("one of validators failed")
        self.validators = validators

    def pred(self, value: T, trace: List[str]) -> bool:
        for validator in self.validators:
            if not validator.pred(value, trace):
                return val_error(trace, "Sequence failed")
        return True


class ListValidator(Validator[List[T]]):

    def __init__(self, validator: Optional[Validator[T]]):
        super().__init__("one of the values is not valid")
        self.validator = validator

    def pred(self, value: List[T], trace: List[str]) -> bool:
        if not isinstance(value, list):
            return val_error(trace, "The value is not list")
        if self.validator is not None:
            for subvalue in value:
                if not self.validator.pred(subvalue, trace):
                    return val_error(trace, "ListValidator failed on "
                                     + f"{subvalue}")
        return True


class OptionalValidator(Validator[T]):

    def __init__(self, validator: Validator[T]):
        super().__init__("the value is not none")
        self.validator = validator

    def pred(self, value: T, trace: List[str]) -> bool:
        if value is None:
            return True
        return self.validator.pred(value, trace)


###############################################################################
# Models
###############################################################################

Kwargs = Any


class Model:
    def __init__(self, collection: Collection):
        self.collection = collection
        self.validator: Optional[Validator[Document]] = None
        self.doc_validator: Optional[Validator[Document]] = None
        self.defaults: Optional[Document] = None

    def build_dict(self, **kwargs: Kwargs) -> Document:
        doc = {}
        for key, value in kwargs.items():
            doc[key] = value
        return doc

    def add(self, **kwargs: Kwargs) -> Optional[Document]:
        doc = self.build_dict(**kwargs)
        return self.d_add(doc)

    def _add_def(self, doc: Document) -> Document:
        if self.defaults is not None:
            for key, value in self.defaults.items():
                if key not in doc:
                    if callable(value):
                        doc[key] = value()
                    else:
                        doc[key] = value
        return doc

    def d_add(self, doc: Document) -> Optional[Document]:
        self._add_def(doc)
        if self.validator is not None and not self.validator.validate(doc):
            return None
        return self.collection.insert_one(doc)

    def d_update(self, find: Document, doc: Document,
                 upsert=True) -> Optional[Document]:

        value = self.collection.find_one_and_update(
            find, {"$set": doc}, upsert=False,
            return_document=ReturnDocument.AFTER)
        if value:
            return value

        if not upsert:
            return None

        return self.d_add({**find, **doc})

    def d_rewrite(self, keys: List[str], doc: Document) -> Optional[Document]:
        find = {}
        for key in keys:
            find[key] = doc[key]
        return self.d_update(find, doc, upsert=True)

    def get(self, **kwargs: Kwargs) -> List[Document]:
        query = self.build_dict(**kwargs)
        return self.collection.find(query)

    def get_one(self, **kwargs: Kwargs) -> Optional[Document]:
        query = self.build_dict(**kwargs)
        return self.collection.find_one(query)

    def get_by_id(self, id: MongoId) -> Optional[Document]:
        return self.collection.find_one({"_id": id})

    def exists(self, **kwargs: Kwargs) -> bool:
        return self.get_one(**kwargs) is not None


class User(Model):

    def __init__(self, collection: Collection):
        super().__init__(collection)
        self.validator = SequentialValidator([
            ReqFieldsValidator(["client_ser", "client_id", "username"]),
            PrimaryKeyValidator(collection, ["client_ser", "client_id"]),
            DictValidator({
                "client_ser": StringValidator(min_size=1, max_size=64),
                "client_id": StringValidator(min_size=1, max_size=64),
                "username": StringValidator(min_size=1, max_size=64),
                "levelcode": OptionalValidator(codeValidator),
                "worldcode": OptionalValidator(codeValidator)
            })
        ])
        self.defaults = {
            "levelcode": None,
            "worldcode": None
        }

    def get_user(self, userkey: UserKey) -> Optional[Document]:
        return self.get_one(client_ser=userkey[0], client_id=userkey[1])

    def get_users(self, userkeys: List[UserKey]) -> List[Document]:
        users = []
        for userkey in userkeys:
            user = self.get_user(userkey)
            if user is not None:
                users.append(user)
        return users

    def add_user(self, userkey: UserKey, username: str) -> Optional[Document]:
        return self.add(client_ser=userkey[0], client_id=userkey[1],
                        username=username)

    def open_user(self, userkey: UserKey, username: str) -> Optional[Document]:
        user = self.get_user(userkey)
        if user:
            return user
        return self.add_user(userkey, username)

    def move_user(self, userkey: UserKey, worldcode: str,
                  levelcode: str) -> Optional[Document]:
        return self.collection.update_one(
            {'client_ser': userkey[0],
             'client_id': userkey[1]},
            {"$set": {"levelcode": levelcode, "worldcode": worldcode}},
            upsert=False
        )


codeValidator = StringValidator(min_size=1, max_size=10)
titleValidator = StringValidator(min_size=1, max_size=64)


class World(Model):

    def __init__(self, collection: Collection):
        super().__init__(collection)
        self.doc_validator = SequentialValidator([
            ReqFieldsValidator(["title", "code", "tags", "prereqs"]),
            DictValidator({
                "title": titleValidator,
                "code": codeValidator,
                "tags": ListValidator(StringValidator(min_size=4,
                                                      max_size=32)),
                "prereqs": ListValidator(RefKeyValidator(collection, "code")),
            })
        ])
        self.validator = SequentialValidator([
            self.doc_validator,
            PrimaryKeyValidator(collection, ["code"])
        ])
        self.defaults = {
            "tags": [],
            "prereqs": [],
        }


class Level(Model):

    def __init__(self, collection: Collection, worlds: Collection):
        super().__init__(collection)
        self.doc_validator = SequentialValidator([
            ReqFieldsValidator(
                ["title", "code", "worldcode", "prereqs", "pipeline"]),
            DictValidator({
                "title": titleValidator,
                "desc": StringValidator(),

                "code": codeValidator,
                "worldcode": RefKeyValidator(worlds, "code"),

                "prereqs": ListValidator(RefKeyValidator(collection, "code")),

                "pipeline": StringValidator(min_size=4, max_size=16),
                "attrs": DictValidator({}),
                "tests": ListValidator(None),

                "tags": ListValidator(StringValidator(min_size=4,
                                                      max_size=32)),
            })
        ])
        self.validator = SequentialValidator([
            PrimaryKeyValidator(collection, ["worldcode", "code"]),
            self.doc_validator
        ])
        self.defaults = {
            "desc": "",
            "tags": [],
            "prereqs": [],
            "tests": [],
            "attrs": {}
        }

    def get_by_code(self, worldcode: str,
                    levelcode: str) -> Optional[Document]:
        return self.get_one(worldcode=worldcode, code=levelcode)


CodeblockValidator = StringValidator(min_size=1, max_size=4096)


class Workplace(Model):

    def __init__(self, collection: Collection, users: Collection,
                 worlds: Collection):
        super().__init__(collection)
        self.doc_validator = SequentialValidator([
            ReqFieldsValidator(["user", "worldcode", "submissions"]),
            DictValidator({
                "user": RefKeyValidator(users, "_id"),
                "worldcode": RefKeyValidator(worlds, "code"),
                "submissions": ListValidator(
                    ListValidator(CodeblockValidator)),
                "bench": ListValidator(CodeblockValidator)
            })
        ])
        self.validator = SequentialValidator([
            PrimaryKeyValidator(collection, ["user", "worldcode"]),
            self.doc_validator
        ])
        self.defaults = {"submissions": [], "bench": []}

    def add_submission(self, userid: MongoId, worldcode: str,
                       submission: List[str]) -> Optional[Document]:
        doc = self.build_dict(user=userid, worldcode=worldcode)
        if self.collection.find_one(doc) is None:
            if self.d_add(doc) is None:
                return None

        document = self.collection.find_one(doc)

        if document is None:
            return None

        submissions = document["submissions"]
        submissions.append(submission)
        if len(submissions) > MAX_CODEBLOCKS:
            submissions = submissions[len(submissions) - MAX_CODEBLOCKS:]
        return self.collection \
            .update_one(doc, {'$set': {"submissions": submissions}})

    def get_workplaces(self, userids: List[MongoId],
                       worldcode: str) -> List[Document]:
        return self.collection.find({"worldcode": worldcode,
                                     "user": {"$in": userids}})

    def get_codeblocks(self, userids: List[MongoId],
                       worldcode: str) -> Dict[MongoId, List[str]]:
        workplaces = self.get_workplaces(userids, worldcode)
        result = {}
        for workplace in workplaces:
            submissions = workplace["submissions"]
            codeblocks = []
            if submissions is not None and len(submissions) != 0:
                codeblocks = submissions[-1]
            result[workplace["user"]] = codeblocks
        return result

    def get_workbenches(self, userids: List[MongoId],
                        worldcode: str) -> Dict[MongoId, List[str]]:
        result = {}
        for workplace in self.get_workplaces(userids, worldcode):
            result[workplace["user"]] = workplace["bench"]
        return result

    def set_workbenches(self, codeblocks: Dict[MongoId, List[str]],
                        worldcode: str):
        results = []
        for userid, blocks in codeblocks.items():
            res = self.d_update(
                {"user": userid, "worldcode": worldcode}, {"bench": blocks})
            print(res)
            if res is not None:
                results.append(res)
        return results


class Solution(Model):

    def __init__(self, collection: Collection, users: Collection,
                 levels: Collection, worlds: Collection):
        super().__init__(collection)
        self.doc_validator = SequentialValidator([
            ReqFieldsValidator(
                ["user", "levelcode", "worldcode", "mark", "date"]),
            DictValidator({
                "user": RefKeyValidator(users, "_id"),
                "levelcode": RefKeyValidator(levels, "code"),
                "worldcode": RefKeyValidator(worlds, "code"),
                "mark": IntegerValidator(0, 3),
                # TODO: date validator
            }),
            RefValidator(levels, {"levelcode": "code",
                                  "worldcode": "worldcode"})
        ])
        self.validator = SequentialValidator([
            PrimaryKeyValidator(
                collection, ['user', 'levelcode', 'worldcode']),
            self.doc_validator
        ])
        self.defaults = {
            "date": datetime.now
        }

    def add_solution(self, userid: MongoId, worldcode: str, levelcode: str,
                     mark: int) -> Optional[Document]:
        sol_doc = self.build_dict(user=userid, levelcode=levelcode,
                                  worldcode=worldcode, mark=mark)
        return self.d_rewrite(['user', 'levelcode', 'worldcode'], sol_doc)

    def get_solutions(self, userid: MongoId, worldcode: Optional[str] = None) \
            -> List[Document]:
        query = {"user": userid}
        if worldcode is not None:
            query['worldcode'] = worldcode
        return self.collection.find(query)


class EllatuDB:

    def __init__(self, host: str, port: Optional[int] = None,
                 db_name: str = DB_DEV):
        self.client = MongoClient(host, port)
        self.db = self.client[db_name]
        self.user = User(self.db[COL_USERS])
        self.world = World(self.db[COL_WORLDS])
        self.level = Level(self.db[COL_LEVELS], self.db[COL_WORLDS])
        self.workplace = Workplace(self.db[COL_WORKPLACES],
                                   self.db[COL_USERS],
                                   self.db[COL_WORLDS])
        self.solution = Solution(self.db[COL_SOLUTIONS],
                                 self.db[COL_USERS],
                                 self.db[COL_LEVELS],
                                 self.db[COL_WORLDS])
